"""
Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

import datetime
import os
import json
import threading
import multiprocessing
import time
import pandas as pd
import webbrowser
import requests
import traceback
from queue import Queue

from wiliot_core import set_logger, InlayTypes
from wiliot_tools.utils.wiliot_gui.wiliot_gui import popup_message
from wiliot_testers.association_tester.modules.performance_module import ReelVerification
from wiliot_testers.association_tester.modules.association_module import TagAssociation
from wiliot_testers.association_tester.modules.gui_module import AssociationAndVerificationGUI
from wiliot_testers.association_tester.modules.configuration_module import RUN_PARAMS_FILE


pd.options.mode.chained_assignment = None  # default='warn'
INLAY_TYPE = InlayTypes.TIKI_121.name


class AssociationAndVerificationTester(object):
    def __init__(self, user_inputs):
        start_time = datetime.datetime.now()
        common_run_name = f"{user_inputs['run_name']}_{start_time.strftime('%Y%m%d_%H%M%S')}"
        self.logger_path, self.logger = set_logger(app_name='AssociationAndVerificationTester',
                                                   common_run_name=common_run_name)
        self.logger_path = os.path.dirname(self.logger_path)
        self.run_data = {'common_run_name': common_run_name,
                         'tester_type': 'association_and_verification',
                         'inlay_type': INLAY_TYPE,
                         'station_name': os.environ['testerStationName']
                         if 'testerStationName' in os.environ else 'testStation',
                         'run_start_time': start_time}
        self.run_data_path = os.path.join(self.logger_path, f'{common_run_name}@run_data.csv')
        self.packets_data_path = os.path.join(self.logger_path, f'{common_run_name}@packets_data.csv')
        self.logger.info(f'run data will be saved at: {self.run_data_path}')
        user_inputs['common_run_name'] = common_run_name
        user_inputs['run_data_path'] = self.run_data_path
        user_inputs['packets_data_path'] = self.packets_data_path
        stop_event = multiprocessing.Event()
        rescan_event = multiprocessing.Event()
        self.gui_done_event = multiprocessing.Event()
        exception_q = Queue(maxsize=100)
        self.app_running = True
        self.performance = None
        try:
            if user_inputs['do_verification']:
                self.performance = ReelVerification(user_input=user_inputs,
                                                    stop_event=stop_event,
                                                    is_app_running=self.is_app_running,
                                                    logger_name=self.logger.name,
                                                    logger_path=self.logger_path,
                                                    exception_q=exception_q)
            self.association = TagAssociation(user_input=user_inputs,
                                            stop_event=stop_event,
                                            rescan_event=rescan_event,
                                            is_app_running=self.is_app_running,
                                            logger_config={'logger_name': self.logger.name, 'logger_path': self.logger_path},
                                            exception_q=exception_q,
                                            run_param_file=RUN_PARAMS_FILE,
                                            )

            self.gui = AssociationAndVerificationGUI(logger_name=self.logger.name,
                                                    get_data_func=self.get_data,
                                                    get_stat_func=self.get_stat,
                                                    stop_event=stop_event,
                                                    tag_association=self.association,
                                                    rescan_event=rescan_event,
                                                    gui_done_event=self.gui_done_event,
                                                    app_config=user_inputs)
        except Exception as e:
            self.logger.warning(f'Got error during init: {e}')
            popup_message(f'Error during initialize the Tester: {e}')
            raise e
        self.user_inputs = user_inputs
        self.stop_event = stop_event
        self.exception_q = exception_q
        self.results_df = pd.DataFrame()
        self.current_stat = {'n_locations': '0',
                             'scan_success': '0%', 'association_success': '0%', 'responding_rate': '0%',
                             'n_success': '0', 'success_rate': '0%'}
        self.duplicated_code = pd.DataFrame()
        self.neglected_duplication = pd.DataFrame()
        self.performance_thread = None
        self.association_thread = None
        self.gui_thread = None

    def run(self):
        if self.user_inputs['do_verification']:
            self.performance_thread = threading.Thread(target=self.performance.run_app, args=())
            self.performance_thread.start()
        
        self.association_thread = threading.Thread(target=self.association.run_app, args=())
        self.association_thread.start()
        
        self.gui_thread = threading.Thread(target=self.gui.run_app, args=())
        self.gui_thread.start()

        webbrowser.open(self.gui.get_url())

        self.run_app()

    def run_app(self):
        while True:
            try:
                time.sleep(1)
                if self.gui.is_stopped_by_user():
                    self.logger.info('run stopped by user')
                    break
                elif self.app_running != self.gui.is_app_running():
                    self.logger.info(f'run was {"paused" if self.app_running else "continued"} by user')
                    self.app_running = not self.app_running

                if not self.exception_q.empty():
                    self.app_running = False
                    self.logger.warning('got exception during run, pause the app')
                    self.handle_exceptions()
                    break  # TODO only stop for some exceptions

                if self.app_running:
                    df_ass = self.association.get_locations_df()
                    df_per = self.get_packets_df()
                    self.merge_results(location_df=df_ass, packets_df=df_per)
                    self.current_stat = self.calc_stat()
                    self.handle_stop_criteria()
                    # save data:
                    self.save_location_data()
                    self.save_run_data()
                else:
                    time.sleep(1)
            except Exception as e:
                self.logger.warning(f'got exception during run: {e}')
                if not self.exception_q.full():
                    self.exception_q.put(f'Main: {e}')
                    self.logger.warning(f'Main: {(traceback.format_exc())}')

        # stop run
        self.stop()

    def get_packets_df(self):
        return self.performance.get_packets_df() if self.performance is not None else pd.DataFrame()

    def is_app_running(self):
        return self.app_running

    def handle_stop_criteria(self):
        if int(self.current_stat['n_successive_bad_scan']) >= int(self.user_inputs['sc_n_no_scan']):
            raise Exception(f"{self.current_stat['n_successive_bad_scan']} successive bad scanning were detected. "
                            f"Check if machine is stuck")

        if self.current_stat['is_duplicated']:
            self.neglected_duplication = self.duplicated_code.copy()
            raise Exception(f"{self.duplicated_code.values} codes duplications were detected. "
                            f"Check if machine is stuck")

        if int(self.current_stat['n_locations']) < int(self.user_inputs['sc_min_location']):
            return
        if self.user_inputs['do_association']:
            if float(self.current_stat['association_success'].replace('%', '')) < float(
                    self.user_inputs['sc_association']):
                self.user_inputs['sc_association'] = 0  # pop only once
                raise Exception(f"association yield of {self.current_stat['association_success']} was detected")
        if float(self.current_stat['scan_success'].replace('%', '')) < float(self.user_inputs['sc_scanning']):
            self.user_inputs['sc_scanning'] = 0  # pop only once
            raise Exception(f"scan yield of {self.current_stat['scan_success']} was detected")
        if self.user_inputs['do_verification']:
            if float(self.current_stat['responding_rate'].replace('%', '')) < float(self.user_inputs['sc_responding']):
                self.user_inputs['sc_responding'] = 0  # pop only once
                raise Exception(f"responding yield of {self.current_stat['responding_rate']} was detected")

    def handle_exceptions(self):
        n_exceptions = self.exception_q.qsize()
        exceptions_str = []
        for _ in range(n_exceptions):
            exceptions_str.append(self.exception_q.get())
        self.logger.warning('\n'.join(exceptions_str))
        popup_message('\n'.join(exceptions_str))

    def print_run(self):  # for debug
        t_i = time.time()
        while time.time() - t_i < 60:
            try:
                time.sleep(2)
                df_ass = self.association.get_locations_df()
                print(f'n unique locations: {len(df_ass)}')
                df_per = self.get_packets_df()
                print(f'n unique adva: {len(df_per)}')
                self.merge_results(location_df=df_ass, packets_df=df_per)
                print(f'n all results: {len(self.results_df)}')
                stat = self.calc_stat()
                print(stat)
            except Exception as e:
                self.logger.warning(f'got exception during print_run: {e}')

        # stop run
        self.stop_event.set()
        self.stop()

    def calc_stat(self):
        stat_out = {}
        if self.results_df.empty:
            stat_out['n_locations'] = '0'
            stat_out['n_tags_outside_test'] = '0'
            stat_out['scan_success'] = '0%'
            stat_out['association_success'] = '0%'
            stat_out['responding_rate'] = '0%'
            stat_out['n_successive_bad_scan'] = 0
            stat_out['is_duplicated'] = False
            stat_out['n_success'] = '0'
            stat_out['success_rate'] = '0%'
            return stat_out
        # calculate the statistics:
        nan_location_index = ((self.results_df['location'].isna()) | (self.results_df['location'] == ''))
        rel_data = self.results_df.loc[~nan_location_index]
        # df for calculation:
        rel_data = rel_data.drop_duplicates(subset=['location', 'asset_code', 'wiliot_code'])
        df_per_loc = rel_data.drop_duplicates(subset=['location'])
        # locations:
        n_location = rel_data['location'].nunique()
        stat_out['n_locations'] = str(n_location)
        stat_out['n_tags_outside_test'] = str(sum(nan_location_index))
        # success rates:
        stat_out['scan_success'] = f'{round(rel_data["scan_status"].sum() / len(rel_data) * 100, 2)}%'  # per wiliot-code
        # association success:
        ass_valid = df_per_loc["is_associated"][df_per_loc["associate_status_code"] != '']
        stat_out['association_success'] = f'{round(ass_valid.sum() / max([ass_valid.count(), 1]) * 100, 2)}%'  # per asset / location
        # responding rate:
        if 'n_packets' in rel_data.keys():
            respond_valid = rel_data["n_packets"][rel_data["wiliot_code"] != '']
            stat_out['responding_rate'] = f'{round(respond_valid.notna().sum() / len(respond_valid) * 100, 2)}%'   # per wiliot-code
        else:
            stat_out['responding_rate'] = '0%'
        # last successive bad scan
        if any(rel_data['scan_status']):
            stat_out['n_successive_bad_scan'] = rel_data['location'].iloc[-1] - rel_data['location'][rel_data['scan_status']].iloc[-1]  # per asset / location
        else:
            stat_out['n_successive_bad_scan'] =  n_location  # per asset / location
        # duplication
        assets_duplication_ind = (df_per_loc.duplicated('asset_code')) & (df_per_loc['asset_code'] != '')  # per asset / location
        pixels_duplication_ind = (rel_data.duplicated(subset=['wiliot_code'])) & (rel_data['wiliot_code'] != '')
        self.duplicated_code = pd.concat([df_per_loc['asset_code'].loc[assets_duplication_ind],rel_data['wiliot_code'].loc[pixels_duplication_ind]], axis=0)
        self.duplicated_code = self.duplicated_code.drop_duplicates()
        stat_out['is_duplicated'] = (len(self.duplicated_code)) > len(self.neglected_duplication)  # per both asset and wiliot code
        # overall success
        if 'is_success' in df_per_loc.keys():
            stat_out['n_success'] = f'{df_per_loc["is_success"].sum()}'   # per asset / location
            stat_out['success_rate'] = f'{round(df_per_loc["is_success"].sum() / n_location * 100, 2)}%'  # per asset / location
        else:
            stat_out['n_success'] = '0'
            stat_out['success_rate'] = '0%'

        return stat_out

    def stop(self):
        self.stop_event.set()
        time.sleep(2)
        while not self.gui_done_event.is_set():
            time.sleep(1)
        time.sleep(4)
        self.gui.server.shutdown()
        all_threads = {'Performance': self.performance_thread, 'Association': self.association_thread, 'Gui': self.gui_thread}
        for t_name, t in all_threads.items():
            if t is not None:
                t.join(15)
                if t.is_alive():
                    self.logger.warning(f'{t_name} thread is still running')

        # summary
        df_ass = self.association.get_locations_df()
        df_per = self.get_packets_df()
        self.merge_results(location_df=df_ass, packets_df=df_per)
        self.current_stat = self.calc_stat()

        self.save_location_data()
        self.save_run_data()

        # handle exceptions:
        if not self.exception_q.empty():
            self.handle_exceptions()

        # show results:
        stat_out_str = "\n".join([f'{k}: {v}' for k, v in self.current_stat.items()])
        if self.user_inputs['do_scan_verification']:
            popup_message('make sure last labels are printed correctly')
        if self.user_inputs['do_association'] and self.is_bad_association_occurred():
            popup_message('Run contains Association Failure, please run and check Association report Manually', bg = 'red')
        popup_message(f'{self.run_data["common_run_name"]}\n\n{stat_out_str}')

    def save_run_data(self):
        data_to_save = pd.DataFrame({**self.run_data, **{'end_run_time': datetime.datetime.now()}, **self.current_stat,
                                     **self.user_inputs}, index=[0])
        data_to_save.to_csv(self.run_data_path, index=False)

    def save_location_data(self):
        if self.results_df.empty:
            return
        data_to_save = self.results_df.loc[~self.results_df['location'].isna()]
        data_to_save.sort_values(by='location', inplace=True)
        data_to_save.to_csv(self.packets_data_path, index=False)

    def is_bad_association_occurred(self):
        return float(self.current_stat['association_success'].replace('%', '')) < 100.0


    def get_data(self):
        return self.results_df

    def get_stat(self):
        return self.current_stat

    def merge_results(self, location_df, packets_df):
        if packets_df.empty and location_df.empty:
            return

        merged_df = None
        if packets_df.empty:
            merged_df = location_df
        else:
            packets_df.index = packets_df.index.set_names(['adv_address'])
            packets_df = packets_df.reset_index()

        if location_df.empty:
            merged_df = packets_df
            merged_df.insert(loc=0, column='location', value='')
            if 'n_packets' not in merged_df.keys():
                merged_df.insert(loc=0, column='n_packets', value=0)

        if merged_df is None:
            merged_df = pd.merge(location_df, packets_df, left_on='wiliot_code', right_on='external_id', how='outer')
        
        if 'n_packets' in merged_df.keys():
            is_responded = merged_df['n_packets'].apply(lambda x: int(x) > 0 if not pd.isnull(x) else False)
        else:
            is_responded = pd.Series([not self.user_inputs['do_verification']] * merged_df.shape[0], index=merged_df.index)
        
        if 'scan_status' not in merged_df.keys():
            merged_df.insert(loc=0, column='scan_status', value=False)
        if 'is_associated' not in merged_df.keys():
            merged_df.insert(loc=0, column='is_associated', value=False)
        
        if self.user_inputs['do_association']:
            is_success = pd.DataFrame([merged_df['scan_status'], merged_df['is_associated'], is_responded]).all()
        else:
            is_success = pd.DataFrame([merged_df['scan_status'], is_responded]).all()
        
        if 'is_success' in merged_df.keys():
            merged_df = merged_df.drop(columns='is_success')
        if 'common_run_name' in merged_df.keys():
            merged_df = merged_df.drop(columns='common_run_name')
        
        merged_df.insert(loc=len(merged_df.columns), column='is_success', value=is_success)
        merged_df.insert(loc=0, column='common_run_name', value=self.run_data['common_run_name'])

        self.results_df = merged_df


if __name__ == '__main__':
    from wiliot_testers.association_tester.modules.configuration_module import get_user_inputs, get_params, GUI_FILE, DEFAULT_VALUES
    user_inputs = get_user_inputs()
    print(user_inputs)
    av = AssociationAndVerificationTester(user_inputs=user_inputs)
    av.run()
    print('done')



"""
Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

import datetime
import numpy as np
import pandas as pd
from pathlib import Path
import time

from wiliot_core.utils.utils import open_directory
from wiliot_testers.failure_analysis_tester.configs_gui import load_config, FA_TEST_DIR, TESTER_NAME, TIME_NOW, RELAY_CONFIG, CONFIGS_DIR, CAPACITANCE_LIMITS_NF
from wiliot_testers.failure_analysis_tester.fa_utils import plot_value, generate_html_report
from wiliot_testers.wiliot_tester_log import WiliotTesterLog
from wiliot_test_equipment.visa_test_equipment.gpio_router import GPIORouter
from wiliot_test_equipment.visa_test_equipment.smu import SMU


class FailureAnalysisTester:
    def __init__(self, debug=False):
        self.set_logger()
        self.df = pd.DataFrame()
        self.all_df = pd.DataFrame()
        self.reference_df = None
        self.load_config()
        self.tag_aliases = []
        self.list_for_html_report_no_plots = []
        self.list_for_html_report_with_plots = []
        self.is_app_running = True
        self.is_first_run = True
        if debug:
            from wiliot_core.internal.utils import get_sharepoint_path
            self.df = pd.read_csv(get_sharepoint_path('Tags Team files') / 'misc' / 'fa_tester.csv')
            self.all_df = pd.read_csv(get_sharepoint_path('Tags Team files') / 'misc' / 'fa_tester_merged.csv')
            self.output_dir = FA_TEST_DIR / 'results' / 'test' / f"test_{TIME_NOW}"
            self.output_dir.mkdir(exist_ok=True, parents=True)
        else:
            self.gpio_router = GPIORouter(logger=self.logger)
            self.smu = SMU(visa_addr=self.config['visa_addr'], logger=self.logger)

    def set_logger(self):
        run_name = 'failure_analysis_' + TIME_NOW
        self.log_obj = WiliotTesterLog(run_name=run_name)
        self.log_obj.set_logger(tester_name=TESTER_NAME,
                                log_path=FA_TEST_DIR / 'logs')
        self.logger = self.log_obj.logger

    def load_config(self):
        self.config = load_config()

    def generate_csv_report(self):
        if self.all_df.empty:
            return
        reshaped_rows = []
        for ch in RELAY_CONFIG.keys():
            if f'{ch}_voltage_V' in self.all_df.columns:
                temp_df = pd.DataFrame({
                    'CH1 Voltage (V)': self.all_df[f'{ch}_voltage_V'],
                    'CH1 Source (uA)': self.all_df[f'{ch}_current_uA'],
                    'ID': self.all_df['tag_alias'],
                    'Target': ch,
                    'result': self.all_df.get(f'{ch}_result', pd.Series([''] * len(self.all_df))),
                })
                reshaped_rows.append(temp_df)
        output_df = pd.concat(reshaped_rows, ignore_index=True)
        output_csv_path = self.output_dir / \
            f'SMU_report_{datetime.datetime.now().strftime("%Y_%m_%d__%H_%M_%S")}.csv'
        output_df.to_csv(output_csv_path, index=False)
        
    def end_test(self):
        self.generate_csv_report()
        self.html_output_path = self.output_dir / f'SMU_report_{datetime.datetime.now().strftime("%Y_%m_%d__%H_%M_%S")}.html'
        main_plot = []
        for ch in RELAY_CONFIG.keys():
            if f'{ch}_voltage_V' in self.all_df.columns:
                main_plot.append(plot_value(self.all_df, self.reference_df, ch))
        generate_html_report(self.list_for_html_report_with_plots,
                             self.html_output_path,
                             main_plot=main_plot)
        open_directory(self.output_dir)
        open_directory(self.html_output_path)

    def load_reference_csv(self, tag_type:str = 'E4'):
        git_config_path = Path(__file__).parent / 'configs'
        csv_files = list(git_config_path.glob(f"*{tag_type}*.csv"))
        if len(csv_files) == 1:
            self.reference_df = pd.read_csv(csv_files[0])

    @staticmethod
    def create_capacitance_chart(cap_nF:float) -> tuple[str, str]:
        if cap_nF is None or np.isnan(cap_nF):
            return "N/A", "N/A"
        half_range = (CAPACITANCE_LIMITS_NF[1] - CAPACITANCE_LIMITS_NF[0]) / 4
        verdict = 'Pass' if CAPACITANCE_LIMITS_NF[0] <= cap_nF <= CAPACITANCE_LIMITS_NF[1] else 'Fail'
        chart_str = "(-----)"
        if verdict == 'Pass':
            idx = np.floor(5 * (cap_nF - CAPACITANCE_LIMITS_NF[0]) / (CAPACITANCE_LIMITS_NF[1] - CAPACITANCE_LIMITS_NF[0])) + 1
            if idx > 5:
                idx = 5
            chart_str = chart_str[:int(idx)] + '*' + chart_str[int(idx)+1:]
        else:
            if cap_nF < CAPACITANCE_LIMITS_NF[0]:
                if cap_nF < CAPACITANCE_LIMITS_NF[0] - half_range:
                    chart_str = "*-" + chart_str
                else:
                    chart_str = "*" + chart_str
            else:
                if cap_nF > CAPACITANCE_LIMITS_NF[1] + half_range:
                    chart_str = chart_str + "-*"
                else:
                    chart_str = chart_str + "*"
        return chart_str, verdict

    def fill_list_for_report(self, tag_alias, test_type:str = 'IV Curve') -> None:
        if test_type == 'Capacitance':
            self.fill_list_for_report_capacitance(tag_alias)
        elif test_type == 'IV Curve':
            self.fill_list_for_report_iv_curve(tag_alias)

    def fill_list_for_report_capacitance(self, tag_alias) -> None:
        chart_str, verdict = self.create_capacitance_chart(self.cap_nF)
        result_dict = {'Test Name': 'Capacitance',
            'Tag Id': tag_alias,
            'Result': f'{self.cap_nF:.2} nF',
            'Limits': f'{CAPACITANCE_LIMITS_NF[0]}-{CAPACITANCE_LIMITS_NF[1]} nF',
            'Result': f'{self.cap_nF:.2} nF',
            'Chart': chart_str,
            'Verdict': verdict,
            'graph': None}
        self.list_for_html_report_no_plots.append(result_dict)

    def fill_list_for_report_iv_curve(self, tag_alias) -> None:
        cols = [s.replace('_current_uA','') for s in list(self.df.columns) if 'current_uA' in s]
        for col in cols:
            percent_pass = self.df.get(col + '_percent_pass', None)
            if percent_pass is not None:
                percent_pass = percent_pass.iloc[0] if isinstance(percent_pass, pd.Series) else percent_pass
                if percent_pass >= 95:
                    result = 'pass'
                else:
                    result = 'fail'
            else:
                percent_pass = 'N/A'
                result = 'N/A'
            
            result_dict = {'Test Name': col,
                           'Tag Id': tag_alias,
                           'Percent Pass': percent_pass,
                           'Result': result,
                           'graph': plot_value(self.df, self.reference_df, col, self.df['tag_alias'][0])}
            self.list_for_html_report_with_plots.append(result_dict)

    def check_test(self, tag_type:str = 'E4', test_type:str = 'IV Curve') -> str:
        if test_type == 'Capacitance':
            chart_str, verdict = FailureAnalysisTester.create_capacitance_chart(self.cap_nF)
            return f'Capacitance is {self.cap_nF:.4} nF\n{verdict}\n{chart_str}'
        elif test_type != 'IV Curve':
            return 'N/A'
        out_str = ''
        if self.reference_df is None:
            self.load_reference_csv(tag_type)

        if self.reference_df is not None:
            thresh = self.config['compare_threshold_percent'] / 100
            cols = [s.replace('_current_uA','') for s in list(self.df.columns) if 'current_uA' in s]
            for col in cols:
                res = 0
                err = False
                for i, (cur, vol) in enumerate(zip(self.df[col + '_current_uA'], self.df[col + '_voltage_V'])):
                    try:
                        nearest_index = self.reference_df[col + '_current_uA'].sub(cur).abs().idxmin()
                        single_res = vol < self.reference_df[col + '_min_voltage'].iloc[nearest_index] or vol > self.reference_df[col + '_max_voltage'].iloc[nearest_index]
                        res += single_res
                        self.df.loc[i, col + '_result'] = 'pass' if single_res else 'fail'
                    except KeyError:
                        out_str += f'{col} is missing\n'
                        err = True
                        break
                if not err:
                    res = 100 * (1 - res / len(self.df[col + '_current_uA']))
                    self.df[col + '_percent_pass'] = res

                    if res >= 95:
                        out_str += f'{col} pass rate: {res:.2f}%  ✅\n'
                    else:
                        out_str += f'{col} pass rate: {res:.2f}%  ❌\n'
                        
                    
            return out_str
        else:
            return 'Reference file issue'

    def run_test(self, test_type: str, folder_name: str, tag_alias: str, comment: str, keys=None):
        # add suffix to tag_alias for retests
        if tag_alias in self.tag_aliases:
            idx = 1
            while tag_alias + '_' + str(idx) in self.tag_aliases:
                idx += 1
            tag_alias = tag_alias + '_' + str(idx)

        self.tag_aliases.append(tag_alias)
        self.logger.info(
            f'running {test_type}, {folder_name}, {tag_alias}')
        if self.is_first_run:
            self.output_dir = FA_TEST_DIR / 'results' / \
                folder_name / f"{folder_name}_{TIME_NOW}"
            self.output_dir.mkdir(exist_ok=True, parents=True)
        csv_path = self.output_dir / \
            f'{test_type}_{tag_alias}_{datetime.datetime.now().strftime("%Y_%m_%d__%H_%M_%S")}.csv'
        if test_type == 'IV Curve':
            self.df = pd.DataFrame()
            self.run_iv_curve_test(keys=keys)
            self.df['tag_alias'] = tag_alias
            self.df['comment'] = comment
            self.df.to_csv(csv_path)
            self.all_df = pd.concat([self.all_df, self.df], ignore_index=True)
            self.all_df.to_csv(self.output_dir / 'merged.csv')
        elif test_type == 'Capacitance':
            self.run_capacitance_test()
        else:
            raise NotImplementedError

    def run_capacitance_test(self):
        self.gpio_router.set_gpio_state(RELAY_CONFIG['VDD_CAP'])

        # charge_and_calc_cap

        self.smu.config_volt_source(channel=1, volt_v=self.config['Capacitance']['clamp_V'], current_limit_mA=self.config['Capacitance']['current_limit_mA'], reset=False)
        time.sleep(0.5)  # let the tag settle

        # ---------- 2.  constant‑current charge ----------
        push_current_mA = self.config['Capacitance']['push_current_uA'] * 1e-3
        self.logger.info(f"[INFO] Configuring SMU for constant current: ~{push_current_mA:.6f} mA, "
              f"measuring voltage for ~{self.config['Capacitance']['measure_time_s']} s.")

        self.smu.config_current_source(
            channel=1,
            current_mA=push_current_mA,
            volt_limit_mV=2000,
            reset=False
        )

        sample_interval_s = 0.05
        num_points = int(self.config['Capacitance']['measure_time_s'] / sample_interval_s)

        self.smu.configure_voltage_sampling(
            channel=1,
            current_mA=push_current_mA,
            volt_limit_V=2.0,
            delay_us=0,
            samp_interval_us=int(sample_interval_s * 1e6),
            num_points=num_points,
            reset=False
        )

        self.logger.info(f"[INFO] Starting sampling: {num_points} points "
              f"at ~{sample_interval_s * 1e3:.1f} ms interval")

        self.smu.run(channel=1)
        time.sleep(self.config['Capacitance']['measure_time_s'])
        t_s, curr_A, volt_V = self.smu.read_sampling(channel=1)
        self.logger.info(f"[INFO] Received {len(t_s)} points from SMU sampling.")

        # ---------- 3.  capacitance calculation ----------
        dv = volt_V[-1] - volt_V[0]
        dt = t_s[-1] - t_s[0]
        I = np.mean(curr_A)  # average current during pulse
        cap_F = (I * dt) / dv if dv > 0 else float('nan')
        self.cap_nF = cap_F * 1e9

        self.logger.info(f"[RESULT]  ΔV={dv:.4f} V,  Δt={dt:.2f} s,  I≈{I * 1e3:.3f} mA  →  C≈{self.cap_nF:.4} nF")

    def run_iv_curve_test(self, keys=None):
        if not keys:
            keys = RELAY_CONFIG.keys()
        for key in keys:
            self.test_field(key)

    def test_field(self, field):
        self.gpio_router.set_gpio_state(RELAY_CONFIG[field])
        kwargs = self.config[field]
        time_list_float, current_A_list_float, voltage_V_list_float = self.smu.run_and_read_current_sweep(
            **kwargs)
        current_uA_list_float = [x * 1e6 for x in current_A_list_float]
        # self.df[field + '_time'] = time_list_float
        df = pd.DataFrame()
        df[field + '_current_uA'] = current_uA_list_float
        df[field + '_voltage_V'] = voltage_V_list_float
        self.df = pd.concat([self.df, df], axis=1)


if __name__ == '__main__':
    FAT = FailureAnalysisTester()
    FAT.output_dir = FA_TEST_DIR / 'results' / 'test' / f"test_{TIME_NOW}"
    FAT.output_dir.mkdir(exist_ok=True, parents=True)
    FAT.check_test('E2')
    FAT.fill_list_for_report()
    FAT.end_test()
    # FAT.run_test('IV Curve', 'smu_fix', 'aaa', '')



"""
Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""
from pathlib import Path
import json

from wiliot_core import GetApiKey
from wiliot_tools.utils.wiliot_gui.wiliot_gui import *
from wiliot_testers.association_tester.hw_components.scanner_component import Scanner
from wiliot_testers.association_tester.hw_components.r2r_component import R2R

GUI_USER_INPUT_PATH = Path(__file__).parents[1]
FILE_NAME = 'association_and_verification_gui_user_inputs.json'
GUI_FILE = GUI_USER_INPUT_PATH / 'configs' / FILE_NAME
RUN_PARAMS_FILE = GUI_USER_INPUT_PATH / 'configs' / 'last_run_params.json'
R2R_PRINTER_PARAMS_FILE = GUI_USER_INPUT_PATH / 'configs' / 'r2r_printer_params.json'
SCANNER_PARAMS_FILE = GUI_USER_INPUT_PATH / 'configs' / 'scanner_params.json'
ASSOCIATION_PARAMS_FILE = GUI_USER_INPUT_PATH / 'configs' / 'association_params.json'
DEFAULT_VALUES = {
    'min_test_time': '1.5',
    'max_test_time': '5',
    'time_to_move': '0.5',
    'do_association': 'no',
    'is_step_machine': 'yes',
    'r2r_type': 'arduino',
    'asset_location': 'none',
    'wiliot_num_codes': 1,
    'asset_num_codes': 1,
    'asset_code_pattern': '',
    'owner_id': 'wiliot-ops',
    'category_id': '86fd9f07-38b0-466f-b717-2e285b803f5c',
    'energy_pattern': '51',
    'time_profile': '5,15',
    'ble_power': '22', 'sub1g_power': '29', 'sub1g_freq': '925000',
    'scan_ch': '37',
    'is_listen_bridge': 'yes',
    'env': 'prod',
    'scanner_type': 'Cognex',
    'rescan_mode': 'none',
    'sc_min_location': '30',
    'sc_association': '90',
    'sc_scanning': '90',
    'sc_responding': '90',
    'sc_n_no_scan': '2',
    'run_name': 'test',
    'do_verification': True,
    'do_printing': False,
    'do_scan_verification': False,
    'scan_verification_offset': 1,
    'pixels_mapping_file': '',
    'pixels_mapping_type': 'none',
    'r2r_print_file': '',
}

DEFAULT_RUN_PARAM = {"last_run_name": "test", "last_location": 0, "last_asset_id": "", "last_asset_location": 0}
DEFAULT_R2R_PRINTER_PARAM = {"printer_name": "", "dpi": 203, "label_format_path": "", "label_content_path": "", "starting_ind": 0, "label_width_in": 4, "label_height_in": 6, "label_gap_in": 0.2}
DEFAULT_SCANNER_PARAM = {"qr_code_mapping": [], "qr_separator": ""}
DEFAULT_ASSOCIATION_PARAM = {"association_field_mapping": {},}
has_r2r_moved = False

def get_params(file_path, default_values):
    if file_path.is_file():
        out = json.load(open(file_path, "rb"))
        for k in default_values.keys():
            if k not in out.keys():
                out[k] = default_values[k]
    else:
        out = default_values
        with open(file_path, 'w') as f:
            json.dump(default_values, f, indent=4)
    return out


def get_user_inputs():
    """
    opens GUI for selecting a file and returns it
    """
    # upload config
    global is_valid_setup_out
    default_values = get_params(GUI_FILE, DEFAULT_VALUES)
    run_params = get_params(RUN_PARAMS_FILE, DEFAULT_RUN_PARAM)
    params_dict = {}
    params_dict['main'] = {

        # Setup Button
        'check_setup': {
            'text': 'Check Setup',
            'value': '',
            'widget_type': 'button',
            'tab': 'Label Location'
        },

        # Label Location Section
        'run_name': {
            'text': 'Run name',
            'value': default_values['run_name'],
            'widget_type': 'entry',
            'tab': 'Label Location'
        },
        'generate_run_name': {
            'text': 'Generate Run Name',
            'value': '',
            'widget_type': 'button',
            'tab': 'Label Location'
        },
        'first_location': {
            'text': 'Start label location',
            'value': 0,
            'widget_type': 'entry',
            'tab': 'Label Location'
        },        
        'last_tested': {
            'value': f'last tested run name was: {run_params["last_run_name"]}\n'
                     f'last tested location was: {run_params["last_location"]}\n'
                     f'last scanned asset id was: {run_params["last_asset_id"]} '
                     f'at location: {run_params["last_asset_location"]}',
            'text': '',
            'widget_type': 'label',
            'tab': 'Label Location'
        },
        'move_r2r': {
            'text': 'Move to the next Label',
            'value': '',
            'widget_type': 'button',
            'tab': 'Label Location'
        },

        # Tag-GW Configuration Section
        'do_verification': {
            'text': 'Do RF Verification',
            'value': default_values['do_verification'],
            'widget_type': 'checkbox',
            'tab': 'Tag-GW Configuration'
        },
        'min_test_time': {
            'text': 'Minimal wait time per location [sec] (for RF test)',
            'value': default_values['min_test_time'],
            'widget_type': 'entry',
            'tab': 'Tag-GW Configuration'
        },
        'energy_pattern': {
            'text': 'Energy Pattern',
            'value': default_values['energy_pattern'],
            'widget_type': 'entry',
            'tab': 'Tag-GW Configuration'
        },
        'time_profile': {
            'text': 'Time Profile',
            'value': default_values['time_profile'],
            'widget_type': 'entry',
            'tab': 'Tag-GW Configuration'
        },
        'scan_ch': {
            'text': 'Scan Channel',
            'value': default_values['scan_ch'],
            'widget_type': 'entry',
            'tab': 'Tag-GW Configuration'
        },
        'ble_power': {
            'text': 'BLE Power[dBm]',
            'value': default_values['ble_power'],
            'widget_type': 'entry',
            'tab': 'Tag-GW Configuration'
        },
        'sub1g_power': {
            'text': 'Sub1G Power[dBm]',
            'value': default_values['sub1g_power'],
            'widget_type': 'entry',
            'tab': 'Tag-GW Configuration'
        },
        'sub1g_freq': {
            'text': 'Sub1G frequency[kHz]',
            'value': default_values['sub1g_freq'],
            'widget_type': 'entry',
            'tab': 'Tag-GW Configuration'
        },
        'is_listen_bridge': {
            'text': 'listen to Bridge?',
            'value': default_values['is_listen_bridge'],
            'options': ('yes', 'no'),
            'widget_type': 'combobox',
            'tab': 'Tag-GW Configuration'
        },

        # Scanner Section
        'scanner_type': {
            'text': 'Scanner Type',
            'value': default_values['scanner_type'],
            'options': ('Cognex', ''),
            'widget_type': 'combobox',
            'group': 'Scanner',
            'tab': 'Hardware'
        },
        'asset_location': {
            'text': 'Asset location with respect to Wiliot code',
            'value': default_values['asset_location'],
            'options': ('first', 'last', 'none'),
            'widget_type': 'combobox',
            'group': 'Scanner',
            'tab': 'Hardware'
        },
        'wiliot_num_codes': {
            'text': 'Number of Wiliot Pixels per label',
            'value': default_values['wiliot_num_codes'],
            'widget_type': 'entry',
            'group': 'Scanner',
            'tab': 'Hardware'
        },
        'asset_num_codes': {
            'text': 'Number of asset codes per label',
            'value': default_values['asset_num_codes'],
            'widget_type': 'entry',
            'group': 'Scanner',
            'tab': 'Hardware'
        },
        'asset_code_pattern': {
            'text': 'Asset code pattern (optional)',
            'value': default_values['asset_code_pattern'],
            'widget_type': 'entry',
            'group': 'Scanner',
            'tab': 'Hardware'
        },
        'max_test_time': {
            'text': 'Maximal wait time per location [sec]',
            'value': default_values['max_test_time'],
            'widget_type': 'entry',
            'group': 'Scanner',
            'tab': 'Hardware'
        },
        'rescan_mode': {
            'text': 'Re-Scan Mode',
            'value': default_values['rescan_mode'],
            'options': ('none', 'asset_only', 'all'),
            'widget_type': 'combobox',
            'group': 'Scanner',
            'tab': 'Hardware'
        },
        'pixels_mapping_type': {
            'text': 'Pixels IDs Rule',
            'value': default_values['pixels_mapping_type'],
            'options': ('none', 'whitelist', 'blacklist'),
            'widget_type': 'combobox',
            'group': 'Scanner',
            'tab': 'Hardware'
        },
        'pixels_mapping_file': {
            'text': 'If Pixels IDs Rule was specified, please provide file path',
            'value': default_values['pixels_mapping_file'],
            'widget_type': 'file_input',
            'group': 'Scanner',
            'tab': 'Hardware'
        },
        'do_printing': {
            'text': 'Print on Failed Labels?',
            'value': default_values['do_printing'],
            'widget_type': 'checkbox',
            'group': 'Printer',
            'tab': 'Hardware'
        },
        'do_scan_verification': {
            'text': 'Scan after printing?',
            'value': default_values['do_scan_verification'],
            'widget_type': 'checkbox',
            'group': 'Verification Scanner',
            'tab': 'Hardware'
        },
        'scan_verification_offset': {
            'text': 'Scanner offset',
            'value': default_values['scan_verification_offset'],
            'options': list(range(1, 10)),
            'widget_type': 'combobox',
            'group': 'Verification Scanner',
            'tab': 'Hardware'
        },

        # Reel-to-Reel Section
        'is_step_machine': {
            'text': 'Is step machine?',
            'value': default_values['is_step_machine'],
            'options': ('yes', 'no'),
            'widget_type': 'combobox',
            'tab': 'Reel-to-Reel'
        },
        'r2r_type': {
            'text': 'step machine type',
            'value': default_values['r2r_type'],
            'options': ('gateway', 'arduino', 'zebra printer'),
            'widget_type': 'combobox',
            'tab': 'Reel-to-Reel'
        },
        'r2r_print_file': {
            'text': 'If using r2r_printer, please choose print file',
            'value': default_values['r2r_print_file'],
            'widget_type': 'file_input',
            'tab': 'Reel-to-Reel'
        },
        'time_to_move': {
            'text': 'Movement time [sec]',
            'value': default_values['time_to_move'],
            'widget_type': 'entry',
            'tab': 'Reel-to-Reel'
        },

        # Association Section
        'do_association': {
            'text': 'Do association?',
            'value': default_values['do_association'] if isinstance(default_values['do_association'], bool) else default_values['do_association'].lower()=='yes',
            'widget_type': 'checkbox',
            'tab': 'Association'
        },
        'owner_id': {
            'text': 'Owner id',
            'value': default_values['owner_id'],
            'widget_type': 'entry',
            'tab': 'Association'
        },
        'env': {
            'text': 'Environment',
            'value': default_values['env'],
            'widget_type': 'entry',
            'tab': 'Association'
        },
        'category_id': {
            'text': 'Asset category id',
            'value': default_values['category_id'],
            'widget_type': 'entry',
            'tab': 'Association'
        },

        # Stop Criteria Section
        'sc_min_location': {
            'text': 'Minimum tags to test before applying "yield-stop-criteria"',
            'value': default_values['sc_min_location'],
            'widget_type': 'entry',
            'tab': 'Stop Criteria'
        },
        'sc_association': {
            'text': 'Stop if association yield is lower than [%]',
            'value': default_values['sc_association'],
            'widget_type': 'entry',
            'tab': 'Stop Criteria'
        },
        'sc_scanning': {
            'text': 'Stop if scanning yield is lower than [%]',
            'value': default_values['sc_scanning'],
            'widget_type': 'entry',
            'tab': 'Stop Criteria'
        },
        'sc_responding': {
            'text': 'Stop if responding yield is lower than [%]',
            'value': default_values['sc_responding'],
            'widget_type': 'entry',
            'tab': 'Stop Criteria'
        },
        'sc_n_no_scan': {
            'text': 'Stop if N successive labels are failed to be scanned',
            'value': default_values['sc_n_no_scan'],
            'widget_type': 'entry',
            'tab': 'Stop Criteria'
        },
    }
    is_valid_setup_out = {'is_valid': False, 'scanner': None, 'scanned_codes': []}

    def on_check_setup():
        global is_valid_setup_out
        values = gui.get_all_values()
        is_valid_setup_out = is_valid_setup(values, is_valid_setup_out['scanner'], gui)

    def on_generate_run_name():
        global is_valid_setup_out
        is_valid_setup_out['scanned_codes'].sort(reverse=True)
        gui.update_widget(widget_key='main_run_name', new_value='_'.join(is_valid_setup_out['scanned_codes']))

    def on_do_verification():
        do_verification = gui.get_all_values()['main_do_verification']
        gw_tab = ['min_test_time', 'energy_pattern', 'time_profile', 'scan_ch', 'ble_power', 'sub1g_power', 'sub1g_freq', 'is_listen_bridge']
        for k in gw_tab:
            gui.update_widget(f'main_{k}', disabled=not do_verification)

    def on_move_r2r():
        global has_r2r_moved
        values = gui.get_all_values()
        try:
            printer_config = get_params(R2R_PRINTER_PARAMS_FILE, DEFAULT_R2R_PRINTER_PARAM)
            printer_config['starting_ind'] = int(values['main_first_location'])
            r2r = R2R(logger_config={'logger_name': 'root', 'logger_path': None}, r2r_type=values['main_r2r_type'],
                      r2r_printer_config=printer_config)
            r2r.move_r2r() # TODO printer needs line selection
            r2r.is_r2r_move(timeout=10)
            r2r.exit_app()
            has_r2r_moved = True
        except Exception as e:
            popup_message(str(e))

    def on_run():
        global is_valid_setup_out
        values = gui.get_all_values()
        if not is_valid_setup_out['is_valid']:
            is_valid_setup_out = is_valid_setup(values, is_valid_setup_out['scanner'], gui)
            if not is_valid_setup_out['is_valid']:
                print('invalid setup - update configuration and run again')
                return
        for k, v in values.items():
            default_values[k] = v
        with open(GUI_FILE, 'w') as f:
            json.dump(default_values, f, indent=4)
        with open(RUN_PARAMS_FILE, 'w') as f:
            json.dump(run_params, f, indent=4)

        if is_valid_setup_out['scanner'] is not None:
            is_valid_setup_out['scanner'].disconnect()
        if values['main_r2r_type'] == 'gateway' and values['main_do_verification']:
            popup_message('Gateway R2R cannot run when running with Verification')
            return default_values
        gui.on_submit()
        return default_values

    gui = WiliotGui(params_dict, do_button_config=False, title='Association and Verification tester',
                    height_offset=50 , disable_all_children_windows=False)
    gui.widgets['main_last_tested'].configure(anchor="w", justify='left')
    gui.add_event(widget_key='main_generate_run_name', command=on_generate_run_name, event_type='button')
    gui.add_event(widget_key='main_check_setup', command=on_check_setup, event_type='button')
    gui.add_event(widget_key='main_do_verification', event_type='button', command=lambda: on_do_verification())
    gui.add_event(widget_key='main_move_r2r', event_type='button', command=lambda: on_move_r2r())
    gui.button_configs(submit_command=on_run)
    on_do_verification()
    gui.run()
    values_out_clean = {k.replace('main_', ''): v for k, v in default_values.items()}
    values_out_clean['category_id'] = values_out_clean['category_id'].strip()
    values_out_clean['scanner_config'] = get_params(SCANNER_PARAMS_FILE, DEFAULT_SCANNER_PARAM)
    values_out_clean['association_config'] = get_params(ASSOCIATION_PARAMS_FILE, DEFAULT_ASSOCIATION_PARAM)
    with open(GUI_FILE, 'w') as f:
        json.dump(values_out_clean, f, indent=4)
    if values_out_clean['r2r_type'] == 'zebra printer':
        values_out_clean['r2r_printer_config'] = get_params(R2R_PRINTER_PARAMS_FILE, DEFAULT_R2R_PRINTER_PARAM)
        if values_out_clean['r2r_print_file']:
            values_out_clean['r2r_printer_config']['label_content_path'] = values_out_clean['r2r_print_file']
        with open(R2R_PRINTER_PARAMS_FILE, 'w') as f:
            json.dump(values_out_clean['r2r_printer_config'], f, indent=4, sort_keys=True)
        values_out_clean['r2r_printer_config']['starting_ind'] = int(values_out_clean['first_location']) + has_r2r_moved
    else:
        values_out_clean['r2r_printer_config'] = None
    return values_out_clean

def is_valid_setup(values, scanner, gui):
    # check scanner
    valid_output = {'is_valid': False, 'scanner': scanner, 'scanned_codes': []}
    if values['main_scanner_type'].lower() == 'cognex':
        try:
            if scanner is None:
                scanner = Scanner(max_test_time=values['main_max_test_time'], 
                                n_codes=int(values['main_asset_num_codes']) + int(values['main_wiliot_num_codes']))
            valid_output['scanned_codes'] = scanner.scan()
            valid_output['scanned_codes'].sort()
            yes_or_no_layout = {
                'question': {
                    'value': f'Starting location was set to: {values["main_first_location"]}\n'
                                f'The scanned codes are:\n{valid_output["scanned_codes"]}\n'
                                f'Is the starting location is correct?\n'
                                f'Are those the codes of the first label?\n\n'
                                f'If not, please click on No, try to re-position scanner and try again',
                    'text': '',
                    'widget_type': 'label',
                },
            }
        except Exception as e:
            popup_message(str(e))
            valid_output['is_valid'] = False
            return valid_output

    def on_no_button():
        print('re-positioning the scanner and try again')
        valid_output['is_valid'] = False
        yes_or_no_gui.on_cancel()

    def on_yes_button():
        client_types = []
        if values['main_do_association']:
            client_types.append('asset')
        if values['main_do_verification']:
            client_types.append(None)
        for client_type in client_types:
            g = GetApiKey(gui_type='ttk',
                            env=values['main_env'],
                            owner_id=values['main_owner_id'],
                            client_type=client_type
                            )
            api_key = g.get_api_key()
            if not api_key:
                file_path = g.get_config_path()
                popup_message(
                    f'Could not found an api key for owner id {values["main_owner_id"]} and env {values["main_env"]}'
                    f'at path: {file_path}', tk_frame=gui.layout)
                valid_output['is_valid'] = False
                yes_or_no_gui.on_submit()
        valid_output['is_valid'] = True
        yes_or_no_gui.on_submit()

    yes_or_no_gui = WiliotGui(params_dict=yes_or_no_layout, parent=gui.layout, do_button_config=False,
                                title='Valid Setup Association and Verification', exit_sys_upon_cancel=False,
                                disable_all_children_windows=False)
    yes_or_no_gui.button_configs(submit_button_text='Yes', submit_command=on_yes_button, cancel_button_text='No', cancel_command=on_no_button)
    yes_or_no_gui.run()
    print(f'is valid setup output: {valid_output}')
    return valid_output

if __name__ == '__main__':
    user_inputs = get_user_inputs()
    print(user_inputs)


"""
  Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

  Redistribution and use of the Software in source and binary forms, with or without modification,
   are permitted provided that the following conditions are met:

     1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     2. Redistributions in binary form, except as used in conjunction with
     Wiliot's Pixel in a product or a Software update for such product, must reproduce
     the above copyright notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the distribution.

     3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
     may be used to endorse or promote products or services derived from this Software,
     without specific prior written permission.

     4. This Software, with or without modification, must only be used in conjunction
     with Wiliot's Pixel or with Wiliot's cloud service.

     5. If any Software is provided in binary form under this license, you must not
     do any of the following:
     (a) modify, adapt, translate, or create a derivative work of the Software; or
     (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
     discover the source code or non-literal aspects (such as the underlying structure,
     sequence, organization, ideas, or algorithms) of the Software.

     6. If you create a derivative work and/or improvement of any Software, you hereby
     irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
     royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
     right and license to reproduce, use, make, have made, import, distribute, sell,
     offer for sale, create derivative works of, modify, translate, publicly perform
     and display, and otherwise commercially exploit such derivative works and improvements
     (as applicable) in conjunction with Wiliot's products and services.

     7. You represent and warrant that you are not a resident of (and will not use the
     Software in) a country that the U.S. government has embargoed for use of the Software,
     nor are you named on the U.S. Treasury Department’s list of Specially Designated
     Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
     You must not transfer, export, re-export, import, re-import or divert the Software
     in violation of any export or re-export control laws and regulations (such as the
     United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
     and use restrictions, all as then in effect

   THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
   OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
   WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
   QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
   IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
   ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
   FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
   (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
   (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
   (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
   (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""
from wiliot_core import WiliotDir

import time
import os
import logging
import datetime
from csv import DictWriter


class WiliotTesterLog(object):
    def __init__(self, run_name=''):
        """
        generates both log files and CSV output files
        :param run_name: the test run name. used as the folder name and the files prefix
        :type run_name: str
        """

        self.stream_handler = None
        self.stream_handler2 = None
        self.stream_handler_gw = None
        self.log_path = ''
        self.file_handler = None
        self.run_name = run_name
        self.logger = None
        self.file_formatter = None
        self.gw_logger = None
        self.results_logger = None
        self.run_data = {}
        self.run_header = []
        self.packets_header = []
        self.run_data_name = ''
        self.packets_data_name = ''
        self.run_data_path = ''
        self.packets_data_path = ''
        self.data_folder = ''
        self.run_start_time = ''
        self.tester_station_name = ''

    def set_logger(self, log_path=None, tester_name='tester'):
        ''' Sets the logger if doesn't exist from main code - INFO level
        3 loggers declared
        self.logger - our main for packet
        self.results_logger - for results
        self.gw_logger - for GW logging'''

        self.logger = logging.getLogger('WiliotLogger')
        formatter = logging.Formatter('\x1b[36;20m%(asctime)s,%(msecs)03d %(name)s %(levelname)s %(message)s',
                                      '%H:%M:%S')
        self.stream_handler = logging.StreamHandler()
        self.stream_handler.setLevel(logging.DEBUG)
        self.stream_handler.setFormatter(formatter)
        self.logger.addHandler(self.stream_handler)
        self.logger.setLevel(logging.DEBUG)
        # Set format - Check PixieLog by Dudan

        self.gw_logger = logging.getLogger('GWLogger')
        self.results_logger = logging.getLogger('Testing')

        formatter2 = logging.Formatter('\x1b[33;20m%(message)s')
        gw_formatter = logging.Formatter('\x1b[38;20m%(asctime)s,%(msecs)03d %(name)s %(levelname)s %(message)s',
                                         '%H:%M:%S')

        self.stream_handler2 = logging.StreamHandler()
        self.stream_handler2.setLevel(logging.DEBUG)
        self.stream_handler2.setFormatter(formatter2)

        self.stream_handler_gw = logging.StreamHandler()
        self.stream_handler_gw.setLevel(logging.INFO)
        self.stream_handler_gw.setFormatter(gw_formatter)

        if log_path is None:
            wiliot_dir = WiliotDir()
            wiliot_dir.create_tester_dir(tester_name)
            self.log_path = os.path.join(wiliot_dir.get_tester_dir(tester_name), 'logs')
        else:
            self.log_path = os.path.join(log_path, self.run_name)

        try:
            if not os.path.isdir(self.log_path):
                os.mkdir(self.log_path)
            logger_name = self.run_name + '.log'
            logger_path = os.path.join(self.log_path, logger_name)
        except Exception as e:
            self.logger.critical('Params input is invalid')
            raise e

        self.file_handler = logging.FileHandler(logger_path, mode='a')
        self.file_formatter = logging.Formatter('%(asctime)s,%(msecs)03d %(name)s %(levelname)s %(message)s', '%H:%M:%S')
        self.file_handler.setFormatter(self.file_formatter)
        self.logger.addHandler(self.file_handler)
        self.results_logger.addHandler(self.file_handler)
        self.gw_logger.addHandler(self.file_handler)

        self.gw_logger.addHandler(self.stream_handler_gw)
        self.gw_logger.setLevel(logging.INFO)

        self.results_logger.addHandler(self.stream_handler2)
        self.results_logger.setLevel(logging.DEBUG)

        self.logger.info(
            'Log created, path : {}'.format(str(self.log_path)))
        time.sleep(0.3)

    def create_data_dir(self, data_path=None, tester_name='tester', run_name='test'):
        """
        create the data directory
        :param data_path: the directory path of the csv data
        :type data_path: str
        :param tester_name: the tester name such as offline teser. relevant only if data path is not specified
        :type tester_name: str
        :param run_name: the test run name. used as the folder name and the files prefix
        :type run_name: str
        :return: the data pth of the run file (configuration for the whole file) and the path of packets file
                 which contains all packets received during test
        :rtype: str, str
        """
        self.run_name = run_name
        if data_path is None:
            wiliot_dir = WiliotDir()
            wiliot_dir.create_tester_dir(tester_name)
            data_path = os.path.join(wiliot_dir.get_tester_dir(tester_name), 'logs')

        # create folders:
        data_folder = os.path.join(data_path, run_name)
        if not os.path.exists(data_folder):
            os.makedirs(data_folder)
        self.data_folder = data_folder
        self.run_data_name = run_name + '@run_data.csv'
        self.packets_data_name = run_name + '@packets_data.csv'
        self.run_data_path = os.path.join(data_folder, self.run_data_name)
        self.packets_data_path = os.path.join(data_folder, self.packets_data_name)

        return self.run_data_path, self.packets_data_path

    def set_new_path(self, path, formatter=None):
        if self.file_handler is not None:
            self.logger.removeHandler(self.file_handler)
            self.results_logger.removeHandler(self.file_handler)
            self.gw_logger.removeHandler(self.file_handler)
        self.file_handler = logging.FileHandler(path, mode='a')
        if formatter is None and self.file_formatter is None:
            self.file_formatter = logging.Formatter('%(asctime)s,%(msecs)03d %(name)s %(levelname)s %(message)s',
                                                    '%H:%M:%S')
        else:
            self.file_formatter = formatter

        self.file_handler.setFormatter(self.file_formatter)
        self.logger.addHandler(self.file_handler)
        self.results_logger.addHandler(self.file_handler)
        self.gw_logger.addHandler(self.file_handler)

    def set_common_run_name(self, reel_name=None, run_start_time=None, lane_str=''):
        if reel_name is None:
            reel_name = 'test'
        if run_start_time is None:
            run_start_time = datetime.datetime.now()
        
        lane_str = lane_str + '_' if lane_str != '' else lane_str
        common_run_name =  reel_name + '_' + lane_str + run_start_time.strftime("%Y%m%d_%H%M%S")

        self.run_name = common_run_name
        self.run_start_time = run_start_time

        return common_run_name

    def set_station_name(self, tester_station_name):
        self.tester_station_name = tester_station_name

    @staticmethod
    def set_console_handler_level(logger, level):
        for handler in logger.handlers:
            original_formatter = handler.formatter  # Store the original formatter
            handler.setLevel(level)
            handler.setFormatter(original_formatter)  # Re-apply the original formatter


def dict_to_csv(dict_in, path, append=False, only_titles=False):
    if append:
        method = 'a'
    else:
        method = 'w'
    with open(path, method, newline='') as f:
        dict_writer = DictWriter(f, fieldnames=dict_in.keys())
        if not append:
            dict_writer.writeheader()
        if not only_titles:
            dict_writer.writerow(dict_in)
        f.flush()
        os.fsync(f.fileno())


from dash import Dash, html, dcc, dash_table, ctx
import dash_bootstrap_components as dbc


def get_rescan_modal(is_r2r_printer: bool):
    return dbc.Modal(
    [
        dbc.ModalHeader("Rescan Window", close_button=False),
        dbc.ModalBody(
            html.Div([
                html.Div([
                    html.H5("Scanner", style={"textAlign": "center", "color": "black", "marginBottom": "10px"}),
                    dbc.Button("Rescan", id="rescan-btn", className="w-100", style={"height": "60px", "background-color": "#0dabb6", "border": "0px"}),
                    dbc.Button("Disconnect Scanner", id="connect-btn", className="w-100", style={"height": "60px", "marginTop": "10px", "background-color": "#0dabb6", "border": "0px"}),
                ], style={
                    "backgroundColor": "#f8f7f4",
                    "color": "white",
                    "width": "240px",
                    "height": "240px",
                    "padding": "12px",
                    "display": "flex",
                    "flexDirection": "column",
                    "justifyContent": "flex-start",
                    "borderRadius": "6px"
                }),
                html.Div([
                    html.H5(children="Printer" if is_r2r_printer else "", style={"textAlign": "center", "color": "black", "marginBottom": "10px"}),
                    dbc.Button("Reprint Last Label", id="reprint-btn", className="w-100", style={"height": "60px", "background-color": "#0dabb6", "border": "0px", "display": "block" if is_r2r_printer else "none"}),
                ], style={
                    "backgroundColor": "#f8f7f4",
                    "color": "white",
                    "width": "240px",
                    "height": "240px",
                    "padding": "12px",
                    "display": "flex",
                    "flexDirection": "column",
                    "justifyContent": "flex-start",
                    "borderRadius": "6px"
                }),
            ], style={"display": "flex", "flexDirection": "row", "justifyContent": "space-between", "gap": "20px"}),
        ),
        dbc.ModalFooter([
            dbc.Textarea(id="scan-results", rows=4, disabled=True, style={"backgroundColor": "#f8f7f4"}),
            html.Div([
                dbc.Button("Stop", id="stop-rescan-modal", n_clicks=0, style={"background-color": "#0dabb6", "border": "0px"}),
                dbc.Button("Continue", id="continue-rescan-modal", n_clicks=0, disabled=False, style={"background-color": "#0dabb6", "border": "0px"}),
            ], className="d-flex justify-content-between w-100 mt-2")
        ])
    ],
    id="rescan-modal",
    is_open=False,  # start closed
    backdrop=False,  # click outside to close
    keyboard=False,  # ESC to close
    centered=True)

def get_upload_modal():
    return dbc.Modal(
    [
        dbc.ModalHeader("Upload Window", close_button=False),
        dbc.ModalBody(
            html.Div([
                html.Label("Environment", style={"textAlign": "center", "color": "black", "marginBottom": "10px"}),
                dcc.Dropdown(id='env-dropdown-upload-modal',options=['prod', 'dev', 'test'],value='prod'),
                html.Label("Owner ID", style={"textAlign": "center", "color": "black", "marginBottom": "10px"}),
                dbc.Input(id="owner-id-upload-modal", type="text", value=''),
            ], style={
                "backgroundColor": "#f8f7f4",
                "color": "white",
                "height": "240px",
                "padding": "12px",
                "display": "flex",
                "flexDirection": "column",
                "justifyContent": "flex-start",
                "borderRadius": "6px"
            }),
        ),
        dbc.ModalFooter([
            dbc.Textarea(id="results-upload-modal", rows=4, disabled=False, style={"backgroundColor": "#f8f7f4"}),
            html.Div([
                dbc.Button("Dismiss", id="dismiss-btn-upload-modal", n_clicks=0, style={"background-color": "#0dabb6", "border": "0px"}),
                dbc.Button("Upload", id="upload-btn-upload-modal", n_clicks=0, style={"background-color": "#0dabb6", "border": "0px"}),
            ], className="d-flex justify-content-between w-100 mt-2")
        ])
    ],
    id="upload-modal",
    is_open=False,  # start closed
    backdrop=False,  # click outside to close
    keyboard=False,  # ESC to close
    centered=True)

if __name__ == "__main__":
    app = Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])
    app.layout = html.Div([get_upload_modal()])
    app.layout.children[0].is_open = True
    app.run(debug=True)



"""
Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""
import time

from wiliot_core import CommandDetails, WiliotGateway, ActionType, DataType
from wiliot_core import Packet, PacketList


MAX_TIME_WITHOUT_PACKETS = 120
MAX_SUB1G_POWER = 29
MAX_BLE_POWER = 22


class AdvaProcess(object):
    """
    Counting the number of unique advas
    """

    def __init__(self, stop_event, logger, adva_process_inputs, packet_and_triggers_q, get_sensors_values, ignore_adva_before_triggers=True):

        self.unq_packets_new_triggers = packet_and_triggers_q
        self.seen_advas = set()
        self.adva_process_inputs = adva_process_inputs
        self.inlay = adva_process_inputs['selected_inlay']
        self.advas_before_tags = set() if ignore_adva_before_triggers else None
        self.stopped_by_user = False
        self.gw_error_connection = False
        self.second_without_packets = False  # This is only used to print no packets after one minute
        self.gw_instance = None
        self.logger = logger
        self.stop = stop_event
        self.last_change_time = 0
        self.n_triggers = 0
        self.needed_time_between_matrices = float(self.adva_process_inputs['user_inputs'].get('time_between_matrices_sec', 0))
        self.get_sensors_value = get_sensors_values

        self.init_gw(self.adva_process_inputs['listener_path'])
        self.gw_reset_config()

    def init_gw(self, listener_path=None):
        try:
            if self.gw_instance is None:
                self.gw_instance = WiliotGateway(auto_connect=True,
                                                 logger_name=self.logger.name,
                                                 log_dir_for_multi_processes=listener_path,
                                                 np_max_packet_in_buffer_before_error=10,
                                                 allow_reset_time=False)

            else:
                # reconnect
                is_connected = self.gw_instance.is_connected()
                if is_connected:
                    self.gw_instance.close_port()
                self.gw_instance.open_port(self.gw_instance.port, self.gw_instance.baud)

            is_connected = self.gw_instance.is_connected()
            if is_connected:
                self.gw_instance.start_continuous_listener()
            else:
                self.logger.warning("Couldn't connect to GW in main thread")
                raise Exception(f"Couldn't connect to GW in main thread")

        except Exception as ee:
            raise Exception(f"Couldn't connect to GW in main thread, error: {ee}")

    def get_gw_version(self):
        if self.gw_instance:
            return self.gw_instance.get_gw_version()[0]
        return ''

    def gw_reset_config(self, start_gw_app=False):
        """
        Configs the gateway
        """
        if self.gw_instance.connected:
            self.gw_instance.reset_gw()
            self.gw_instance.reset_listener()
            time.sleep(2)
            if not self.gw_instance.is_gw_alive():
                self.logger.warning('gw_reset_and_config: gw did not respond')
                raise Exception('gw_reset_and_config: gw did not respond after rest')

            gw_config = self.inlay

            cmds = {CommandDetails.scan_ch: gw_config['received_channel'],
                    CommandDetails.time_profile: gw_config['time_profile_val'],
                    CommandDetails.set_energizing_pattern: gw_config['energy_pattern_val'],
                    CommandDetails.set_sub_1_ghz_power: [MAX_SUB1G_POWER],
                    CommandDetails.set_scan_radio: self.gw_instance.get_cmd_symbol_params(
                        freq_str=gw_config['symbol_val']),
                    CommandDetails.set_rssi_th: self.adva_process_inputs['user_inputs'].get('rssi_threshold', 0),
                    }
            output_power_cmds = self.gw_instance.get_cmds_for_abs_output_power(abs_output_power=MAX_BLE_POWER)
            cmds = {**cmds, **output_power_cmds}
            self.gw_instance.set_configuration(cmds=cmds, start_gw_app=False, read_max_time=1)
            pin_num = self.adva_process_inputs['user_inputs'].get('pin_number')
            cmd = '!cmd_gpio CONTROL_IN P%s 0' % pin_num.zfill(3)
            self.gw_instance.write(cmd, must_get_ack=True)
            user_commands = self.adva_process_inputs['user_inputs'].get('gw_commands', [])
            for cmd in user_commands:
                self.gw_instance.write(cmd, must_get_ack=True)
            # start GW applicaion if needed
            self.gw_instance.set_configuration(start_gw_app=start_gw_app)
        else:
            raise Exception('Could NOT connect to GW')

    def set_stopped_by_user(self, stopped):
        self.stopped_by_user = stopped

    def check_new_tags(self):
        """
        Does the process of adding a new Packet to our collection
        """
        new_packet_list = PacketList()
        if not self.gw_instance.is_data_available():
            return new_packet_list, False
        
        raw_packets_in = self.gw_instance.get_packets(action_type=ActionType.ALL_SAMPLE, data_type=DataType.RAW,
                                                      tag_inlay=self.inlay['inlay'])
        for p in raw_packets_in:
            cur_p = Packet(p['raw'], time_from_start=p['time'],
                           inlay_type=self.inlay['inlay'])

            tag_id = cur_p.get_adva()

            if self.n_triggers == 0 and self.advas_before_tags is not None:
                self.advas_before_tags.add(tag_id)
                self.logger.info(f'ignore the following adva: {tag_id} before triggering. Currently {len(self.advas_before_tags)} tags were ignored')

            elif tag_id not in self.seen_advas and (self.advas_before_tags is None or tag_id not in self.advas_before_tags):
                self.seen_advas.add(tag_id)
                self.logger.info(f"New adva {tag_id}")
                new_packet_list.append(cur_p)

        return new_packet_list, len(raw_packets_in) > 0

    def check_new_trigger(self):
        triggers_out = []
        if not self.gw_instance.is_signals_available():
            return triggers_out
        
        gw_rsps = self.gw_instance.get_gw_signals()
        for gw_rsp in gw_rsps:
            if not gw_rsp:
                continue
            time_condition_met = (gw_rsp['time'] - self.last_change_time >= self.needed_time_between_matrices) or self.last_change_time == 0
            if time_condition_met:
                if 'Detected High-to-Low peak' in gw_rsp['raw']:
                    self.last_change_time = gw_rsp['time']
                    self.n_triggers += 1
                    self.logger.info(f'Got a Trigger.  Number of Triggers {self.n_triggers}')
                    triggers_out.append({'trigger_time': self.last_change_time, 
                                         'trigger_num': self.n_triggers})
        return triggers_out

    def gw_recovery_flow(self):
        try:
            if not self.gw_instance.is_connected():
                self.logger.info('Trying to reconnect to GW')
                self.init_gw()
            self.gw_reset_config(start_gw_app=True)
            self.gw_error_connection = False
        except Exception as e:
            self.gw_error_connection = True
            self.logger.warning(f"Couldn't reconnect GW, due to: {e}")


    def run(self):
        """
        Receives available data then counts and returns the number of unique advas.
        """
        self.gw_instance.set_configuration(start_gw_app=True)
        self.gw_instance.reset_start_time()
        is_running = True
        last_packet_time = time.time()
        while not self.stop.is_set():
            time.sleep(0)
            try:
                # check gw errors:
                if self.gw_instance.get_read_error_status() or not self.gw_instance.is_connected():
                    self.gw_recovery_flow()
                #  Pause / Start cases
                if not self.stopped_by_user and not is_running:  # user click on Start case
                    self.gw_reset_config(start_gw_app=True)
                    is_running = True
                    last_packet_time = time.time()
                elif self.stopped_by_user and is_running:  # user click on Stop
                    self.gw_instance.reset_gw()
                    is_running = False
                    last_packet_time = time.time()
                
                if not is_running:
                    continue
                #  Receiving triggers
                trigger_list = self.check_new_trigger()

                #  Receiving packets
                new_packet_list, got_packets = self.check_new_tags()
                new_sensors = {}

                if got_packets:
                    # received packets
                    last_packet_time = time.time()

                if len(new_packet_list) > 0:
                    # received new packets
                    new_sensors = self.get_sensors_value()
                    trigger_list += self.check_new_trigger()  # add triggers before and after packets
                elif time.time() - last_packet_time > MAX_TIME_WITHOUT_PACKETS:
                    # Not receiving packets for awhile
                    self.logger.warning(f'Gateway did not get packets for the last {MAX_TIME_WITHOUT_PACKETS} seconds, try to reset and re-config')
                    self.gw_recovery_flow()
                    last_packet_time = time.time()
                else:
                    # Not receiving packets
                    time.sleep(0.050)
                # add to queue:
                if trigger_list or new_packet_list:
                    if self.unq_packets_new_triggers.full():
                        element = self.unq_packets_new_triggers.get()
                        element['packet'] = f'n packets in list: {len(element["packet"])}'
                        self.logger.warning(f"unq_packets_new_triggers is full, dropping {element}")
                    self.unq_packets_new_triggers.put({'trigger': trigger_list, 'packet': new_packet_list, 'sensors': new_sensors})
            except Exception as e:
                self.logger.warning(f'got exception during adva process run: {e}')
        
        self.logger.info('Stop advaProcess run')
        self.gw_instance.reset_gw()
        self.gw_instance.exit_gw_api()



"""
Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

import logging
import os
import multiprocessing
import time
import pandas as pd
import numpy as np
import traceback

from wiliot_core import WiliotGateway, ActionType, DataType, set_logger
from wiliot_tools.resolver_tool.resolve_packets import ResolvePackets, TagStatus
from wiliot_testers.utils.wiliot_external_ids import is_external_id_valid
from wiliot_testers.association_tester.configs.packets_configs import get_packet_param
pd.options.mode.chained_assignment = None  # default='warn'

RESET_GW_TIME = 300  # number of seconds we allow to not getting packets until we reset the GW
RESOLVE_Q_SIZE = 100


class ResolverProcess(ResolvePackets):
    def __init__(self, owner_id, env, need_to_resolve_q, resolved_q, stop_event):
        logger_path, self.logger = set_logger('ResolverProcess', 'resolver_process', 'resolver')
        self.resolved_q = resolved_q
        super().__init__(tags_in_test=[], owner_id=owner_id, env=env, resolve_q=need_to_resolve_q,
                         set_tags_status_df=self.add_to_resolved_q, stop_event_trig=stop_event,
                         logger_name=self.logger.name)
        self.run()

    def check_tag_status(self, ex_id):
        if is_external_id_valid(ex_id['externalId']):
            status = TagStatus.INSIDE_TEST
        else:
            status = TagStatus.OUT_INVALID
            self.logger.warning(f'found invalid external id: {ex_id["externalId"]}, check serialization')
        self.logger.info(f'found new tag in test: {ex_id}')
        return status

    def add_to_resolved_q(self, element):
        self.resolved_q.put(element, block=False)


class ReelVerification(object):
    def __init__(self, user_input, stop_event, is_app_running, logger_name, logger_path, exception_q):
        self.logger = logging.getLogger(logger_name)
        self.logger_path = logger_path
        self.user_inputs = user_input
        self.stop_event = stop_event
        self.exception_q = exception_q
        self.is_running = False
        self.is_main_app_running = is_app_running

        # connect to the gw:
        self.gw_obj = None
        self.gw_init()
        self.gw_reset_and_config()

        # init data:
        self.packets_df = pd.DataFrame()
        # start app
        try:
            need_to_resolve_q = multiprocessing.Queue(maxsize=RESOLVE_Q_SIZE)
            resolved_q = multiprocessing.Queue(maxsize=RESOLVE_Q_SIZE)
            self.resolver_handler = multiprocessing.Process(target=ResolverProcess,
                                                            args=(self.user_inputs['owner_id'],
                                                                  self.user_inputs['env'],
                                                                  need_to_resolve_q,
                                                                  resolved_q,
                                                                  stop_event))

            self.need_to_resolve_q = need_to_resolve_q
            self.resolved_q = resolved_q

        except Exception as e:
            self.logger.warning(f'exception during ReelVerification init: {e}')
            raise e

    def gw_init(self):
        """
        initialize gw and the data listener thread
        """
        self.gw_obj = WiliotGateway(auto_connect=True, logger_name=self.logger.name,
                                    log_dir_for_multi_processes=self.logger_path)
        try:
            if self.gw_obj.connected:
                self.gw_obj.start_continuous_listener()
            else:
                raise Exception('gateway was not detected, please check connection')
        except Exception as e:
            self.logger.warning(e)
            raise e

    def gw_reset_and_config(self):
        """
        config gw to tester mode
        """
        self.gw_obj.reset_gw()
        if not self.gw_obj.is_gw_alive():
            raise Exception('GW is not responding')

        energy_pattern = int(self.user_inputs['energy_pattern'])
        time_profile = [int(x) for x in self.user_inputs['time_profile'].split(',')]
        ble_power = int(self.user_inputs['ble_power'])
        sub1g_power = int(self.user_inputs['sub1g_power'])
        sub1g_freq = int(self.user_inputs['sub1g_freq'])
        scan_ch = int(self.user_inputs['scan_ch'])
        is_bridge = self.user_inputs['is_listen_bridge'].lower() == 'yes'
        self.gw_obj.start_continuous_listener()
        if is_bridge:
            self.gw_obj.write('!listen_to_tag_only 0', with_ack=True)
        self.gw_obj.write(f'!set_sub_1_ghz_energizing_frequency {sub1g_freq}', with_ack=True)
        self.gw_obj.config_gw(received_channel=scan_ch,
                              energy_pattern_val=energy_pattern,
                              time_profile_val=time_profile,
                              effective_output_power_val=ble_power,
                              sub1g_output_power_val=sub1g_power,
                              start_gw_app=True, with_ack=True)

    def run_app(self):
        try:
            self.resolver_handler.start()
        except Exception as e:
            self.logger.warning(f'exception during ReelVerification run_app: {e}')
            raise e
        self.run()

    def run(self):
        self.logger.info('GetPackets Start')
        self.is_running = True
        cur_time = time.time()
        while True:
            time.sleep(0.050)
            try:
                if self.stop_event.is_set():
                    self.logger.info('GetPackets Stop')
                    self.exit_app()
                    return
                elif self.is_running != self.is_main_app_running():
                    self.is_running = not self.is_running
                    if self.is_running:
                        self.continue_app()
                    else:
                        self.pause_app()
                elif not self.exception_q.empty():
                    time.sleep(0.1)
                    continue

                if self.is_running:
                    # check if there is data to read
                    if self.gw_obj.is_data_available():
                        cur_time = time.time()
                        tag_collections_in = self.gw_obj.get_packets(action_type=ActionType.ALL_SAMPLE,
                                                                     data_type=DataType.TAG_COLLECTION)
                        for tag in tag_collections_in.keys():
                            if tag in self.packets_df.index:
                                self.packets_df.loc[tag, 'n_packets'] += 1
                            else:
                                self.logger.info(f'new tag was detected: {tag}')
                                try:
                                    new_packet = tag_collections_in[tag][0]
                                    self.need_to_resolve_q.put({'tag': tag,
                                                                'payload': new_packet.get_payload()},
                                                               block=False)
                                    self.packets_df = pd.concat([self.packets_df,
                                                                 pd.DataFrame({'n_packets': 1,
                                                                               'external_id': 'unknown'},
                                                                              index=[tag])])
                                    self.update_selected_packet(new_packet, tag)
                                except Exception as e:
                                    self.logger.warning(f'could not add new data to need_to_resolve_q or to packets_df '
                                                        f'due to: {e}')
                            self.get_one_packet(tag, tag_collections_in[tag])

                    elif self.gw_obj.get_read_error_status():
                        raise Exception(f'GW got serial reading error during run. Please disconnect and reconnect the GW')

                    if not self.resolved_q.empty():
                        self.merge_resolved_data()

                    if time.time() - cur_time > RESET_GW_TIME:
                        self.logger.warning(f'did NOT get any packets after {RESET_GW_TIME} seconds. reset and config gw')
                        self.gw_reset_and_config()
                        cur_time = time.time()
                else:
                    time.sleep(1)
                    cur_time = time.time()

            except Exception as e:
                self.logger.warning(f'GetPackets got exception: {e}, traceback: {traceback.format_exc()}')
                if not self.exception_q.full():
                    self.exception_q.put(f'GetPackets: {e}')

    def update_selected_packet(self, new_packet, tag, ind=0):
        self.packets_df.loc[tag, 'encrypted_packet'] = new_packet.get_packet_content(get_gw_data=True,
                                                                                     sprinkler_num=ind)
        self.packets_df.loc[tag, 'rssi'] = np.take(new_packet.gw_data['rssi'], ind)
        self.packets_df.loc[tag, 'packet_type'] = new_packet.packet_data['decrypted_packet_type']

    def get_one_packet(self, tag, packet_list):

        for p in packet_list:
            desired_packet_type = get_packet_param(p.packet_data['flow_ver'], 'packet_type')
            if p.packet_data['decrypted_packet_type'] != desired_packet_type:
                continue
            if self.packets_df.loc[tag, 'packet_type'] != desired_packet_type:
                self.update_selected_packet(p, tag)

            for ind in range(len(p)):
                rssi = np.take(p.gw_data['rssi'], ind)
                if rssi < self.packets_df.loc[tag, 'rssi']:
                    self.update_selected_packet(p, tag, ind)

    def merge_resolved_data(self):
        n = self.resolved_q.qsize()
        for _ in range(n):
            resolved = self.resolved_q.get(block=False)
            self.packets_df.loc[resolved['adv_address'], 'external_id'] = resolved['external_id']

    def pause_app(self):
        self.gw_obj.reset_gw()
        time.sleep(1)
        self.gw_obj.exit_gw_api()

    def continue_app(self):
        self.gw_init()
        self.gw_reset_and_config()

    def exit_app(self):
        try:
            self.gw_obj.reset_gw()
            time.sleep(1)
            self.gw_obj.exit_gw_api()
            self.resolver_handler.join(10)
            if self.resolver_handler.is_alive():
                self.logger.warning('resolver process is still running')
            if not self.resolved_q.empty():
                self.merge_resolved_data()
        except Exception as e:
            self.logger.warning(f'GetPackets: exit_app: got exception: {e}')
            if not self.exception_q.full():
                self.exception_q.put(f'GetPackets: exit_app: {e}')

    def get_packets_df(self):
        return self.packets_df


if __name__ == '__main__':
    from wiliot_core import set_logger
    RUN_TIME = 60

    tag_pref_logger_path, tag_pref_logger = set_logger(app_name='ReelVerification', dir_name='reel_verification',
                                                       file_name='verification_log')
    stop_event = multiprocessing.Event()
    user_input = {
        'energy_pattern': '51',
        'time_profile': '5,15',
        'ble_power': '22', 'sub1g_power': '29', 'sub1g_freq': '925000',
        'scan_ch': '37',
        'is_listen_bridge': 'yes',
        'owner_id': 'wiliot-ops',
        'env': 'prod'
    }
    v = ReelVerification(user_input=user_input, stop_event=stop_event,
                         logger_name=tag_pref_logger.name, logger_path=os.path.dirname(tag_pref_logger_path))

    t_i = time.time()
    while time.time() - t_i < RUN_TIME:
        time.sleep(1)
        df = v.get_packets_df()
        print(f'n unique adva: {len(df)}')
    # stop run
    stop_event.set()

    df = v.get_packets_df()
    df.index = df.index.set_names(['adv_address'])
    df = df.reset_index()
    df_path = tag_pref_logger_path.replace('.log', '_packets_df.csv')
    print(f'saving data at: {df_path}')
    df.to_csv(df_path, index=False)

    print('done')



from jinja2 import Template
import numpy as np
import pandas as pd
from pathlib import Path
import plotly.io as pio
import plotly.graph_objects as go
from typing import Union


def plot_value(df: pd.DataFrame, reference_df: Union[pd.DataFrame, None], value: str, tag_alias: str = '') -> go.Figure:
    def denan(arr):
        """Remove NaN values from an array."""
        return arr[~np.isnan(arr)]
    fig = go.Figure()
    for tag in df['tag_alias'].unique():
        p_df = df[df['tag_alias'] == tag]
        fig.add_trace(go.Scatter(
            x=p_df[value + '_current_uA'].values, 
            y=p_df[value + '_voltage_V'].values, 
            mode='markers', 
            name=str(tag),
            marker=dict(size=6)
        ))

    if reference_df is not None and value + '_current_uA' in reference_df.columns:
        x_ref = denan(reference_df[value + '_current_uA'].values)
        lower_limit = denan(reference_df[value + '_min_voltage'].values)
        upper_limit = denan(reference_df[value + '_max_voltage'].values)
        
        # Add red dashed lines for control limits
        fig.add_trace(go.Scatter(
            x=x_ref, 
            y=lower_limit, 
            mode='lines+markers',
            name="Min Control Limit",
            line=dict(color='red', dash='dash', width=2),
            marker=dict(color='red', size=4)
        ))
        fig.add_trace(go.Scatter(
            x=x_ref, 
            y=upper_limit, 
            mode='lines+markers',
            name="Max Control Limit",
            line=dict(color='red', dash='dash', width=2),
            marker=dict(color='red', size=4)
        ))
        
        # Add shaded area between control limits
        fig.add_trace(go.Scatter(
            x=x_ref,
            y=upper_limit,
            mode='lines',
            line=dict(width=0),
            showlegend=False,
            hoverinfo='skip'
        ))
        fig.add_trace(go.Scatter(
            x=x_ref,
            y=lower_limit,
            mode='lines',
            line=dict(width=0),
            fill='tonexty',
            fillcolor='rgba(255, 0, 0, 0.2)',
            showlegend=False,
            hoverinfo='skip'
        ))
    
    chart_title = f"{tag_alias}_{value}" if tag_alias else f"Test_{value}"
    
    fig.update_layout(
        title=chart_title,
        showlegend=True,
        xaxis_title="Current [uA]", 
        yaxis_title="Voltage [V]",
        legend=dict(
            x=0.02,
            y=0.98,
            bgcolor='rgba(255, 255, 255, 0.8)',
            bordercolor='black',
            borderwidth=1
        )
    )

    return fig


def generate_html_report(list_for_report:list, output_path:Path, main_plot:list):
    """
    Generate an HTML report from a list of test result dictionaries.
    The keys are extracted from the dictionary, the only required key is:
    - 'graph': plotly.graph_objects.Figure | None
    """
    # Convert plotly figures to HTML snippets
    for item in list_for_report:
        item['graph_html'] = pio.to_html(item['graph'], include_plotlyjs=True, full_html=False) if item['graph'] is not None else None

    main_plot_html = []
    for p in main_plot:
        main_plot_html.append(pio.to_html(p, include_plotlyjs=True, full_html=False))

    # Jinja2 HTML template
    template = Template("""
    <html>
    <head>
        <meta charset="utf-8">
        <title>Test Report</title>
        <script src="https://cdn.plot.ly/plotly-latest.min.js"></script>
    </head>
    <body>
        <h1>Test Report</h1>
        <table border="1" cellpadding="10" style="border-collapse: collapse; width: 100%;">
            <tr>
                {% for col in cols %}
                <th>{{col}}</th>
                {% endfor %}
            </tr>
            {% for row in results %}
            <tr>
                {% for col in cols %}
                <td>{{ row[col] }}</td>
                {% endfor %}
            </tr>
            {% if row['graph_html'] is not none %}
            <tr>
                <td colspan="4">{{ row['graph_html'] | safe }}</td>
            </tr>
            {% endif %}
            {% endfor %}
        </table>
        {% if main_plot_html %}
        <h1>All Results</h1>
        <table border="1" cellpadding="10" style="border-collapse: collapse; width: 100%;">
            {% for p in main_plot_html %}
            <tr>
                <td colspan="4">{{ p | safe }}</td>
            </tr>
            {% endfor %}
        {% endif %}
        </table>
    </body>
    </html>
    """)

    # Render and save
    cols = list(list_for_report[0].keys())
    if 'graph' in cols: cols.remove('graph')
    if 'graph_html' in cols: cols.remove('graph_html')
    html_output = template.render(results=list_for_report, cols=cols, main_plot_html=main_plot_html)
    with open(output_path, 'w', encoding='utf-8') as f:
        f.write(html_output)

if __name__ == '__main__':
    l = []
    l.append({'Test Name': 'col','Tag Id': 'tag_alias','percent_pass': 54.3214235,'result': 'N/A','graph': None})
    l.append({'Test Name': 'col','Tag Id': 'tag_alias','percent_pass': None,'result': 'N/A','graph': None})
    generate_html_report(l, Path('/Users/eilam.sher/Downloads/report/report.html'),[])



"""
Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

from shutil import copyfile
from os import makedirs
try:
    import pygubu
    from tkinter import Tk, Toplevel
    from os.path import join, abspath, dirname, isfile, isdir
except Exception as e:
    print(f'could not import tkinter or pygubu: {e}')
from json import dump, load
from wiliot_core import WiliotDir
from wiliot_testers.utils.get_version import get_version

TESTER_NAME = 'sample'
VER = get_version()
MEASURED = {'button': 'Measured', 'label': '[meter]'}
MANUAL = {'button': 'Manual', 'label': '[db]'}
DEFAULT_EMULATED_SURFACE = 'no simulation'

wiliot_dir = WiliotDir()
wiliot_dir.create_tester_dir(TESTER_NAME)
sample_test_dir = wiliot_dir.get_tester_dir(TESTER_NAME)
OUTPUT_DIR = abspath(join(sample_test_dir, 'results'))
LOGS_DIR = abspath(join(sample_test_dir, 'logs'))
CONFIGS_DIR = abspath(join(sample_test_dir, 'configs'))
if not isdir(abspath(join(OUTPUT_DIR))):
    makedirs(abspath(join(OUTPUT_DIR)))
if not isdir(abspath(join(LOGS_DIR))):
    makedirs(abspath(join(LOGS_DIR)))
if not isdir(abspath(join(CONFIGS_DIR))):
    makedirs(abspath(join(CONFIGS_DIR)))


class ConfigsGui(object):
    is_gui = False
    atten_cur_mode = MANUAL
    test_config = ''
    config = ''
    configs_test = {}
    paramDict = {}
    defaultConfigsDict = {}
    default_surfaces = {}
    all_configs_fields = []
    builder = None
    ttk = None
    mainwindow = None
    
    def __init__(self, top_builder=None, tk_frame=None):
        self.top_builder = top_builder
        self.tk_frame = tk_frame
        self.config_path = abspath(join(dirname(__file__), 'configs', '.default_test_configs.json'))
        if isfile(self.config_path):
            with open(self.config_path, 'r') as jsonFile:
                self.configs_test = load(jsonFile)
                if len(self.configs_test):
                    self.all_configs_fields = list(self.configs_test[next(iter(self.configs_test))].keys())
            self.fix_antenna_type()

        with open(abspath(join(dirname(__file__), 'configs', '.default_surfaces.json')), 'r') as jsonFile:
            self.default_surfaces = load(jsonFile)

        tests_suites_path = abspath(join(dirname(__file__), 'configs', 'tests_suites.json'))
        if isfile(tests_suites_path):
            with open(tests_suites_path, 'r') as jsonFile:
                tests_suites = load(jsonFile)
                for k in tests_suites.keys():
                    test_time = 0
                    for t in tests_suites[k]['tests']:
                        test_time += (t.get('maxTime', 0) + t.get('delayBeforeNextTest', 0))
                    tests_suites[k] = {**tests_suites[k], **{'isTestSuite': True, 'testTime': test_time}}
                self.configs_test = {**self.configs_test, **tests_suites}

        self.copy_config_file()

    def copy_config_file(self):
        copyfile(self.config_path, abspath(join(CONFIGS_DIR, f'.default_test_configs(ViewOnly)_{VER}.json')))

    def gui(self):
        self.builder = builder = pygubu.Builder()
        ui_file = abspath(join(abspath(dirname(__file__)), 'utils', 'configs.ui'))
        self.builder.add_from_file(ui_file)
        
        img_path = abspath(join(abspath(dirname(__file__)), ''))
        builder.add_resource_path(img_path)
        img_path = abspath(join(abspath(dirname(__file__)), 'utils'))
        builder.add_resource_path(img_path)

        if self.tk_frame is not None:
            self.ttk = Toplevel(self.tk_frame)
        else:
            self.ttk = Tk()

        self.ttk.title("Sample Test Configs")
        
        self.mainwindow = self.builder.get_object('mainwindow', self.ttk)
        
        self.builder.connect_callbacks(self)
        
        self.ttk.protocol("WM_DELETE_WINDOW", self.close_cb)
        self.ttk.lift()
        self.ttk.attributes("-topmost", True)
        self.ttk.attributes("-topmost", False)
        
        self.set_gui_defaults()
        
        self.is_gui = True
        self.ttk.mainloop()
    
    def set_gui_defaults(self):
        temp_dict = self.configs_test
        self.builder.get_object('configsList')['values'] = [key for key, item in temp_dict.items()
                                                            if isinstance(item, dict) and not item.get('isTestSuite', False)]
        if temp_dict.get(self.config) and not temp_dict[self.config].get('isTestSuite', False):
            self.builder.get_object('configsList').set(self.config)
            self.top_builder.tkvariables.get('testTime').set(temp_dict[self.config]['testTime'])
            self.builder.get_object('EmulateSurface')['values'] = tuple(
                self.default_surfaces['EmulateSurface'].keys())
            self.builder.get_object('EmulateSurface').set(DEFAULT_EMULATED_SURFACE)
            self.builder.get_object('antennaType')['values'] = ['TEO', 'TIKI']
            for param, value in temp_dict[self.config].items():
                if self.builder.tkvariables.get(param) is not None:
                    self.builder.tkvariables.get(param).set(value)
                else:
                    pass
        self.builder.get_object('save')['state'] = 'normal'
    
    def fix_antenna_type(self):
        tempDict = self.configs_test
        for config, params in tempDict.items():
            if 'antennaType' in params.keys() and params['antennaType']:
                fixedAntenna = 'TIKI' if params['antennaType'].lower() in ['dual', 'tiki'] else 'TEO'
                self.configs_test[config]['antennaType'] = fixedAntenna

    def is_gui_opened(self):
        return self.is_gui
    
    def get_params(self):
        return self.paramDict
    
    def get_configs(self):
        return self.configs_test
    
    def set_params(self, test_config):
        if test_config not in self.configs_test.keys():
            config_options = list(self.configs_test.keys())
            test_config = config_options[0]
        self.paramDict = self.configs_test[test_config].copy()
        if 'EmulateSurface' in self.configs_test[test_config].keys():
            surface = self.configs_test[test_config]['EmulateSurface']
            self.paramDict['EmulateSurfaceValue'] = self.configs_test[test_config]['EmulateSurfaceValue'] = \
                self.default_surfaces['EmulateSurface'][surface]
        self.top_builder.tkvariables.get('testTime').set(self.paramDict['testTime'])
    
    def set_default_config(self, test_config):
        self.test_config = test_config
        self.config = test_config

    def config_set(self, config):
        self.config = config
        self.set_params(config)
        if self.is_gui:
            self.builder.get_object('configsList').set(config)
            self.set_gui_defaults()
        # except BaseException:
        #     pass

    # ############## GUI Callbacks  #######################

    def choose_antenna_type_cb(self, *args):
        antenna = self.builder.tkvariables.get('antennaType').get()
        antenna = 'TIKI' if antenna.lower() in ['dual', 'tiki'] else 'TEO'
        self.builder.tkvariables.get('antennaType').set(antenna)

    def reset_cb(self):
        self.configs_test.update(self.defaultConfigsDict)
        self.set_gui_defaults()

    def atten_mode_cb(self):
        self.atten_cur_mode = MEASURED if self.atten_cur_mode == MANUAL else MANUAL
        self.builder.tkvariables.get('atten_mode').set(self.atten_cur_mode['button'])
        ble_label = self.builder.tkvariables.get('attenBleLabel')
        ble_label.set(ble_label.get().split()[0] + ' ' + self.atten_cur_mode['label'])
        lora_label = self.builder.tkvariables.get('attenLoRaLabel')
        lora_label.set(lora_label.get().split()[0] + ' ' + self.atten_cur_mode['label'])

    def close_cb(self):
        self.is_gui = False
        self.ttk.destroy()

    def save_cb(self):
        self.config = config = self.builder.get_object('configsList').get()
        if config == '':
            return

        if config not in self.configs_test.keys():
            self.configs_test[config] = {}
        
        if self.configs_test[config].get('isTestSuite', False):
            print('please edit test suite parameters using the json file directly')
            return
        
        for param in self.all_configs_fields:
            value = self.builder.tkvariables.get(param)
            if value is not None:
                self.configs_test[config][param] = value.get()
        self.builder.get_object('configsList')['values'] = [key for key, item in self.configs_test.items()
                                                            if isinstance(item, dict)]
        self.top_builder.get_object('test_config')['values'] = [key for key, item in self.configs_test.items()
                                                                if isinstance(item, dict)]
        self.top_builder.get_object('test_config').set(config)

        with open(self.config_path, 'w+') as jsonFile:
            dump(self.configs_test, jsonFile, indent=4)
        self.copy_config_file()

        self.set_params(config)
        print(f'{config} configuration saved successfully.')

    def config_select_cb(self, *args):
        self.config = config = self.builder.get_object('configsList').get()
        self.builder.get_object('save')['state'] = 'normal'
        if config and self.configs_test[config].get('isTestSuite', False):
            self.builder.get_object('save')['state'] = 'disabled'
        self.top_builder.get_object('test_config').set(config)
        self.reset_cb()
